using UnityEngine;
using System;
using System.Collections.Generic;

namespace FrogFrog.MPAC
{
    public class MPAC : MVRScript
    {
        private Dictionary<string, Anchors> dancerAnchors = new Dictionary<string, Anchors>();
        private GameObject cameraRoot = new GameObject("CameraRoot");
        private GameObject cameraTrans = new GameObject("CameraTrans");
        private GameObject cameraBase = new GameObject("CameraBase");
        private GameObject focusCenter = new GameObject("FocusCenter");
        private GameObject detector = new GameObject("Detector");
        private GameObject distance100 = new GameObject("Distance100");
        private FreeControllerV3 thisAtomObject;
        private float positionDeltaFactor = -1.0F;
        private float rotationDeltaFactor = -1.0F;
        private float focusPositionFactor = -1.0F;
        private float lastFrameTime = Time.fixedDeltaTime;
        private List<string> activeDancer = new List<string>();
        private string currentDancer = "";
        private string currentFocusDancer = "";
        private bool loading = true;
        private float timeDancer = 0.0F;
        private float timeFocus = 0.0F;
        private AutoFloat swingH = new AutoFloat();
        private AutoFloat swingV = new AutoFloat();
        private AutoFloat axisY = new AutoFloat();
        private AutoFloat distance = new AutoFloat();
        private V3Fellow cRootPositionV3F = new V3Fellow();
        private V3Fellow cBasePositionV3F = new V3Fellow();
        private V3Fellow cRootRotationV3F = new V3Fellow() { isAngle = true, halfRound = true, rateX = 0.25F, rateY = 0.5F };
        private V3Fellow cBaseRotationV3F = new V3Fellow() { isAngle = true };
        private V3Fellow focusPositionV3F = new V3Fellow();
        public System.Random randomObj = new System.Random();

        private List<JSONStorableStringChooser> jsDancerList = new List<JSONStorableStringChooser>();
        private JSONStorableFloat jsPositionFollowingStrength = new JSONStorableFloat("Position Following Strength", 0.6F, 0.0F, 1.0F);
        private JSONStorableFloat jsRotationFollowingStrength = new JSONStorableFloat("Rotation Following Strength", 0.6F, 0.0F, 1.0F);
        private JSONStorableFloat jsHorzRotationFollowingRate = new JSONStorableFloat("Horizontal Rotation Following Rate", 0.5F, 0.0F, 1.0F);
        private JSONStorableFloat jsVertRotationFollowingRate = new JSONStorableFloat("Vertical Rotation Following Rate", 0.25F, 0.0F, 1.0F);
        private JSONStorableFloat jsHorzRandomSwingRange = new JSONStorableFloat("Horizontal Random Swing Range", 40.0F, 0.0F, 60.0F);
        private JSONStorableFloat jsVertRandomSwingRange = new JSONStorableFloat("Vertical Random Swing Range", 20.0F, 0.0F, 60.0F);
        private JSONStorableFloat jsRelativeCameraHeight = new JSONStorableFloat("Relative Camera Height", 0.2F, -0.5F, 0.5F);
        private JSONStorableFloat jsAxisYRandomOffsetRange = new JSONStorableFloat("Axis Y Random Offset Range", 0.2F, 0.0F, 0.5F);
        private JSONStorableFloat jsRelativeCameraDistance = new JSONStorableFloat("Relative Camera Distance", 2.0F, 1.5F, 5.0F);
        private JSONStorableFloat jsDistanceRandomOffsetRange = new JSONStorableFloat("Distance Random Offset Rang", 0.5F, 0.0F, 1.0F);
        private JSONStorableFloat jsCameraPitchAngle = new JSONStorableFloat("Camera Pitch Angle", 7.5F, -30.0F, 30.0F);
        private JSONStorableFloat jsRandomMotionTimeMin = new JSONStorableFloat("Random Motion Time (Min)", 4.0F, 1.0F, 30.0F);
        private JSONStorableFloat jsRandomMotionTimeMax = new JSONStorableFloat("Random Motion Time (Max)", 8.0F, 1.0F, 30.0F);
        private JSONStorableFloat jsDancerSwitchTimeMin = new JSONStorableFloat("Dancer Switch Time (Min)", 7.5F, 1.0F, 30.0F);
        private JSONStorableFloat jsDancerSwitchTimeMax = new JSONStorableFloat("Dancer Switch Time (Max)", 15.0F, 1.0F, 30.0F);
        private JSONStorableBool jsAutoCameraOn = new JSONStorableBool("Auto Camera On", false);
        private JSONStorableBool jsVertScreenFocusOptimize = new JSONStorableBool("Vertical Screen Focus Optimize", false);
        private JSONStorableString jsSelectedDancer = new JSONStorableString("Selected Dancer", "");

        public override void Init()
        {
            jsAutoCameraOn.setCallbackFunction = val => { if (val) RefreshActiveDancer(); };
            RegisterBool(jsAutoCameraOn);
            CreateToggle(jsAutoCameraOn);

            RegisterBool(jsVertScreenFocusOptimize);
            CreateToggle(jsVertScreenFocusOptimize);

            List<string> dancerNames = new List<string> { "<None>" };
            List<Atom> allAtoms = SuperController.singleton.GetAtoms();
            foreach (var oneAtom in allAtoms) { if (oneAtom.type == "Person") { dancerNames.Add(oneAtom.uid); } }
            int i = 0;
            jsDancerList.Clear();
            foreach (var dancer in dancerNames)
            {
                if (dancer != "<None>")
                {
                    jsDancerList.Add(new JSONStorableStringChooser($"Dancer{i}", dancerNames, "<None>", $"Dancer{i}"));
                    jsDancerList[i].setCallbackFunction = val => { CheckDancerName(val); };
                    RegisterStringChooser(jsDancerList[i]);
                    CreateScrollablePopup(jsDancerList[i]);
                    i++;

                    Atom dancerAtom = SuperController.singleton.GetAtomByUid(dancer);
                    dancerAnchors.Add(dancer, new Anchors
                    {
                        head = new GameObject(dancer + "_head"),
                        chest = new GameObject(dancer + "_chest"),
                        hipF = new GameObject(dancer + "_hipF"),
                        hipB = new GameObject(dancer + "_hipB"),
                        dancerHead = dancerAtom.GetStorableByID("headControl") as FreeControllerV3,
                        dancerChest = dancerAtom.GetStorableByID("chestControl") as FreeControllerV3,
                        dancerHip = dancerAtom.GetStorableByID("hipControl") as FreeControllerV3,
                        dancerRoot = dancerAtom.GetStorableByID("control") as FreeControllerV3
                    });

                    //head anchor setting
                    dancerAnchors[dancer].head.transform.position = new Vector3(0, 0, 0);
                    dancerAnchors[dancer].head.transform.rotation = Quaternion.identity;
                    dancerAnchors[dancer].head.transform.SetParent(dancerAnchors[dancer].dancerHead.transform, false);
                    dancerAnchors[dancer].head.transform.localPosition = new Vector3(0.0F, 0.0F, 0.1F);

                    //chest anchor setting
                    dancerAnchors[dancer].chest.transform.position = new Vector3(0, 0, 0);
                    dancerAnchors[dancer].chest.transform.rotation = Quaternion.identity;
                    dancerAnchors[dancer].chest.transform.SetParent(dancerAnchors[dancer].dancerChest.transform, false);
                    dancerAnchors[dancer].chest.transform.localPosition = new Vector3(0.0F, 0.1F, 0.15F);

                    //hipF anchor setting
                    dancerAnchors[dancer].hipF.transform.position = new Vector3(0, 0, 0);
                    dancerAnchors[dancer].hipF.transform.rotation = Quaternion.identity;
                    dancerAnchors[dancer].hipF.transform.SetParent(dancerAnchors[dancer].dancerHip.transform, false);
                    dancerAnchors[dancer].hipF.transform.localPosition = new Vector3(0.0F, -0.1F, 0.1F);

                    //hipB anchor setting
                    dancerAnchors[dancer].hipB.transform.position = new Vector3(0, 0, 0);
                    dancerAnchors[dancer].hipB.transform.rotation = Quaternion.identity;
                    dancerAnchors[dancer].hipB.transform.SetParent(dancerAnchors[dancer].dancerHip.transform, false);
                    dancerAnchors[dancer].hipB.transform.localPosition = new Vector3(0.0F, -0.1F, -0.15F);
                    dancerAnchors[dancer].hipB.transform.localEulerAngles = new Vector3(0.0F, 180.0F, 0.0F);
                }
            }

            jsPositionFollowingStrength.setCallbackFunction = val => { positionDeltaFactor = CalculateFactor(val); };
            RegisterFloat(jsPositionFollowingStrength);
            CreateSlider(jsPositionFollowingStrength, true);

            jsRotationFollowingStrength.setCallbackFunction = val => { rotationDeltaFactor = CalculateFactor(val); };
            RegisterFloat(jsRotationFollowingStrength);
            CreateSlider(jsRotationFollowingStrength, true);

            jsHorzRotationFollowingRate.setCallbackFunction = val => { cRootRotationV3F.rateY = val; };
            RegisterFloat(jsHorzRotationFollowingRate);
            CreateSlider(jsHorzRotationFollowingRate, true);

            jsVertRotationFollowingRate.setCallbackFunction = val => { cRootRotationV3F.rateX = val; };
            RegisterFloat(jsVertRotationFollowingRate);
            CreateSlider(jsVertRotationFollowingRate, true);

            RegisterFloat(jsHorzRandomSwingRange);
            CreateSlider(jsHorzRandomSwingRange, true);

            RegisterFloat(jsVertRandomSwingRange);
            CreateSlider(jsVertRandomSwingRange, true);

            RegisterFloat(jsRelativeCameraHeight);
            CreateSlider(jsRelativeCameraHeight, true);

            RegisterFloat(jsAxisYRandomOffsetRange);
            CreateSlider(jsAxisYRandomOffsetRange, true);

            RegisterFloat(jsRelativeCameraDistance);
            CreateSlider(jsRelativeCameraDistance, true);

            RegisterFloat(jsDistanceRandomOffsetRange);
            CreateSlider(jsDistanceRandomOffsetRange, true);

            RegisterFloat(jsCameraPitchAngle);
            CreateSlider(jsCameraPitchAngle, true);

            RegisterFloat(jsRandomMotionTimeMin);
            CreateSlider(jsRandomMotionTimeMin, true);

            RegisterFloat(jsRandomMotionTimeMax);
            CreateSlider(jsRandomMotionTimeMax, true);

            RegisterFloat(jsDancerSwitchTimeMin);
            CreateSlider(jsDancerSwitchTimeMin, true);

            RegisterFloat(jsDancerSwitchTimeMax);
            CreateSlider(jsDancerSwitchTimeMax, true);

            RegisterString(jsSelectedDancer);

            cameraRoot.transform.position = new Vector3(0, 0, 0);
            cameraRoot.transform.rotation = Quaternion.identity;
            cameraRoot.transform.eulerAngles = new Vector3(0, 0, 0);
            cRootPositionV3F.SetThisV3(cameraRoot.transform.position);
            cRootRotationV3F.SetThisV3(cameraRoot.transform.eulerAngles);

            cameraTrans.transform.position = new Vector3(0, 0, 0);
            cameraTrans.transform.rotation = Quaternion.identity;
            cameraTrans.transform.eulerAngles = new Vector3(0, 180, 0);
            cameraTrans.transform.SetParent(cameraRoot.transform);

            cameraBase.transform.position = new Vector3(0, 0, 0);
            cameraBase.transform.rotation = Quaternion.identity;
            cameraBase.transform.eulerAngles = new Vector3(0, 180, 0);
            cBasePositionV3F.SetThisV3(cameraBase.transform.position);
            cBaseRotationV3F.SetThisV3(cameraBase.transform.eulerAngles);

            focusCenter.transform.position = new Vector3(0, 0, 0);
            focusCenter.transform.rotation = Quaternion.identity;
            focusCenter.transform.eulerAngles = new Vector3(0, 0, 0);
            focusCenter.transform.SetParent(cameraBase.transform, false);

            detector.transform.position = new Vector3(0, 0, 0);
            detector.transform.rotation = Quaternion.identity;
            detector.transform.eulerAngles = new Vector3(0, 0, 0);
            detector.transform.SetParent(cameraBase.transform, false);

            thisAtomObject = containingAtom.GetStorableByID("control") as FreeControllerV3;
        }

        private void CheckDancerName(string val)
        {
            int k = 0;
            int i = 0;
            foreach (var dancer in jsDancerList)
            {
                if (val != "<None>" && dancer.val == val)
                {
                    if (k > 0) { jsDancerList[i].val = "<None>"; }
                    k++;
                }
                i++;
            }
        }

        private void OnDestroy()
        {
            foreach (var kvp in dancerAnchors)
            {
                Destroy(kvp.Value.head, 0.0f);
                Destroy(kvp.Value.chest, 0.0f);
                Destroy(kvp.Value.hipF, 0.0f);
                Destroy(kvp.Value.hipB, 0.0f);
            }
            Destroy(focusCenter, 0.0f);
            Destroy(detector, 0.0f);
            Destroy(cameraBase, 0.0f);
            Destroy(cameraTrans, 0.0f);
            Destroy(cameraRoot, 0.0f);
            //SuperController.LogMessage("GameObjects Are Dectroyed");
        }

        private void FixedUpdate()
        {
            if (SuperController.singleton.isLoading) return;
            if (loading) { jsAutoCameraOn.val = false; loading = false; }
            if (!jsAutoCameraOn.val) return;
            if (activeDancer.Count < 1) return;
            if (jsVertScreenFocusOptimize.val)
            {
                float focusOffset = 360 - SuperController.singleton.monitorCameraFOV / 10;
                focusCenter.transform.localEulerAngles = new Vector3(focusOffset, 0.0F, 0.0F);
            }
            else { focusCenter.transform.localEulerAngles = new Vector3(0.0F, 0.0F, 0.0F); }

            bool forceUpdate = false;
            if (lastFrameTime != Time.fixedDeltaTime)
            {
                forceUpdate = true;
                lastFrameTime = Time.fixedDeltaTime;
            }
            if (positionDeltaFactor < 0 || forceUpdate) positionDeltaFactor = CalculateFactor(jsPositionFollowingStrength.val);
            if (rotationDeltaFactor < 0 || forceUpdate) rotationDeltaFactor = CalculateFactor(jsRotationFollowingStrength.val);
            if (focusPositionFactor < 0 || forceUpdate) focusPositionFactor = CalculateFactor(0.95F);

            if (activeDancer.Contains(jsSelectedDancer.val) && jsSelectedDancer.val != currentDancer)
            {
                SetCurrentDancer(jsSelectedDancer.val);
            }
            else if (currentDancer == "" || !activeDancer.Contains(currentDancer))
            {
                SetCurrentDancer(activeDancer[0]);
            }
            else if (timeDancer <= 0)
            {
                if (activeDancer.Count > 1)
                {
                    float totalRight = 0.0F;
                    float thisRight = 0.0F;
                    foreach (var KVP in dancerAnchors)
                    {
                        if (activeDancer.Contains(KVP.Key) && KVP.Key != currentDancer) { totalRight += dancerAnchors[KVP.Key].rank; }
                    }
                    float randomF = Mathf.Clamp(totalRight * (float)randomObj.NextDouble(), 0, totalRight);
                    foreach (var KVP in dancerAnchors)
                    {
                        if (activeDancer.Contains(KVP.Key) && KVP.Key != currentDancer)
                        {
                            thisRight += dancerAnchors[KVP.Key].rank;
                            if (thisRight >= randomF)
                            {
                                SetCurrentDancer(KVP.Key);
                                break;
                            }
                        }
                    }
                }
                else SetCurrentDancer(currentDancer);
            }
            else
            {
                if (dancerAnchors[currentDancer].rank < 1) timeDancer -= Time.fixedDeltaTime / dancerAnchors[currentDancer].rank;
                else timeDancer -= Time.fixedDeltaTime;
            }

            if (swingH.isFiniched) { swingH.SetNewAuto(jsHorzRandomSwingRange.val, jsRandomMotionTimeMin.val, jsRandomMotionTimeMax.val); }
            float swingHOffset = swingH.GetNext();

            if (swingV.isFiniched) { swingV.SetNewAuto(jsVertRandomSwingRange.val, jsRandomMotionTimeMin.val, jsRandomMotionTimeMax.val); }
            float swingVOffset = swingV.GetNext();

            if (axisY.isFiniched) { axisY.SetNewAuto(jsAxisYRandomOffsetRange.val, jsRandomMotionTimeMin.val, jsRandomMotionTimeMax.val); }
            float axisYOffset = axisY.GetNext();

            if (distance.isFiniched) { distance.SetNewAuto(jsDistanceRandomOffsetRange.val, jsRandomMotionTimeMin.val, jsRandomMotionTimeMax.val); }
            float distanceOffset = distance.GetNext();

            Vector3 cRootP = dancerAnchors[currentDancer].chest.transform.position;
            Vector3 direction = dancerAnchors[currentDancer].chest.transform.up;
            direction.Normalize();
            cRootP += direction * axisYOffset;
            cRootPositionV3F.factor = positionDeltaFactor;
            cameraRoot.transform.position = cRootPositionV3F.GetNewV3(cRootP);

            Vector3 cRootR = dancerAnchors[currentDancer].chest.transform.eulerAngles;
            cRootRotationV3F.factor = rotationDeltaFactor;
            cRootR = cRootRotationV3F.NormalizeTarget(cRootR);
            cRootR.x += swingVOffset;
            cRootR.y += swingHOffset;
            cRootR.z = 0.0F;
            cameraRoot.transform.eulerAngles = cRootRotationV3F.GetNewVector3(cRootR);

            cameraTrans.transform.localPosition = new Vector3(0.0F, jsRelativeCameraHeight.val, distanceOffset + jsRelativeCameraDistance.val);
            //cameraTrans.transform.localEulerAngles = new Vector3(jsCameraPitchAngle.val % 360, 180.0F, 0.0F);

            float cHightLimit = dancerAnchors[currentDancer].dancerRoot.transform.position.y + 0.2F - cameraTrans.transform.position.y;
            if (cHightLimit > 0)
            {
                float adjustAngle = Mathf.Atan(cHightLimit / cameraTrans.transform.localPosition.z) * 180.0F / Mathf.PI;
                Vector3 adjustV3 = cameraRoot.transform.eulerAngles;
                adjustV3.x -= adjustAngle;
                adjustV3.x %= 360;
                if (adjustV3.x > 90 && adjustV3.x < 270)
                {
                    if (adjustV3.x <= 180) adjustV3.x = 90;
                    else adjustV3.x = 270;
                }
                cameraRoot.transform.eulerAngles = adjustV3;
            }

            cBasePositionV3F.factor = positionDeltaFactor;
            cameraBase.transform.position = cBasePositionV3F.GetNewV3(cameraTrans.transform.position);

            cBaseRotationV3F.factor = rotationDeltaFactor;
            Vector3 tmpV3 = cameraTrans.transform.eulerAngles;
            tmpV3.x = (tmpV3.x + jsCameraPitchAngle.val) % 360;
            cameraBase.transform.eulerAngles = cBaseRotationV3F.GetNewV3(tmpV3);

            thisAtomObject.transform.position = cameraBase.transform.position;
            thisAtomObject.transform.eulerAngles = cameraBase.transform.eulerAngles;

            float fof = SuperController.singleton.monitorCameraFOV * 0.3F;
            float lastRank = 0.0F;
            float thisRank = 0.0F;
            string thisFocusDancer = "";
            Vector3 FPV3 = new Vector3(0.0F, 0.0F, 0.0F);

            float distMin = 9999999.0F;
            float distMax = 0.0F;

            foreach (var KVP in dancerAnchors)
            {
                if (activeDancer.Contains(KVP.Key))
                {
                    thisRank = GetFocusRank(KVP.Value.head, fof) * 1.25F;
                    //SuperController.LogMessage($"HEAD: {thisRank}");
                    if (thisRank > lastRank)
                    {
                        FPV3 = KVP.Value.head.transform.position;
                        lastRank = thisRank;
                        thisFocusDancer = KVP.Key;
                    }
                    thisRank = GetFocusRank(KVP.Value.chest, fof);
                    //SuperController.LogMessage($"CHEST: {thisRank}");
                    if (thisRank > lastRank)
                    {
                        FPV3 = KVP.Value.chest.transform.position;
                        lastRank = thisRank;
                        thisFocusDancer = KVP.Key;
                    }
                    thisRank = GetFocusRank(KVP.Value.hipF, fof);
                    //SuperController.LogMessage($"HIPF: {thisRank}");
                    if (thisRank > lastRank)
                    {
                        FPV3 = KVP.Value.hipF.transform.position;
                        lastRank = thisRank;
                        thisFocusDancer = KVP.Key;
                    }
                    thisRank = GetFocusRank(KVP.Value.hipB, fof);
                    //SuperController.LogMessage($"HIPB: {thisRank}");
                    if (thisRank > lastRank)
                    {
                        FPV3 = KVP.Value.hipB.transform.position;
                        lastRank = thisRank;
                        thisFocusDancer = KVP.Key;
                    }

                    dancerAnchors[KVP.Key].dist = Vector3.Distance(KVP.Value.chest.transform.position, distance100.transform.position);
                    if (distMin > dancerAnchors[KVP.Key].dist) distMin = dancerAnchors[KVP.Key].dist;
                    if (distMax < dancerAnchors[KVP.Key].dist) distMax = dancerAnchors[KVP.Key].dist;
                }
            }

            if (currentFocusDancer != thisFocusDancer)
            {
                currentFocusDancer = thisFocusDancer;
                timeFocus = 0.0F;
            }
            if (currentDancer == currentFocusDancer) timeFocus = 0.0F;
            else if (currentFocusDancer != "")
            {
                timeFocus += Time.fixedDeltaTime;
                if (timeFocus > 2.5F) { SetCurrentDancer(currentFocusDancer); timeFocus = 0.0F; }
            }

            bool AFPisFound = true;
            FreeControllerV3 AutoFocusPoint = new FreeControllerV3();
            try { AutoFocusPoint = SuperController.singleton.GetAtomByUid("AutoFocusPoint").GetStorableByID("control") as FreeControllerV3; }
            catch (System.Exception) { SuperController.LogError("Can not find 'AutoFocusPoint' atom"); AFPisFound = false; }

            if (AFPisFound)
            {
                focusPositionV3F.factor = focusPositionFactor;
                AutoFocusPoint.transform.position = focusPositionV3F.GetNewV3(FPV3);
            }

            foreach (var KVP in dancerAnchors)
            {
                if (activeDancer.Contains(KVP.Key))
                {
                    dancerAnchors[KVP.Key].rank = 1.0F;
                    if (distMax - 0.333F < dancerAnchors[KVP.Key].dist) dancerAnchors[KVP.Key].rank = 0.6667F;
                    if (distMin + 0.333F > dancerAnchors[KVP.Key].dist) dancerAnchors[KVP.Key].rank = 1.3333F;
                }
            }
        }

        private float CalculateFactor(float strength)
        {
            if (strength > 0.9999999F) return 1.0F;
            return Mathf.Clamp((Mathf.Pow((1 / (1 - Mathf.Sqrt(strength))), Time.fixedDeltaTime) - 1), 0.0F, 1.0F);
        }

        private void RefreshActiveDancer()
        {
            activeDancer.Clear();
            foreach (var oneDancer in jsDancerList) { if (oneDancer.val != "<None>") activeDancer.Add(oneDancer.val); }
        }

        private void SetCurrentDancer(string dancer)
        {
            currentDancer = dancer;
            jsSelectedDancer.val = dancer;
            timeDancer = (jsDancerSwitchTimeMax.val - jsDancerSwitchTimeMin.val) * (float)randomObj.NextDouble() + jsDancerSwitchTimeMin.val;
            cRootRotationV3F.SetRefV3(dancerAnchors[dancer].dancerRoot.transform.eulerAngles);
            distance100.transform.position = dancerAnchors[dancer].dancerRoot.transform.position;
            Vector3 dir = dancerAnchors[dancer].dancerRoot.transform.forward;
            dir.Normalize();
            distance100.transform.position += dir * 100;
        }

        private float GetFocusRank(GameObject target, float fof)
        {
            float faceToAngle = (target.transform.eulerAngles.y - cameraBase.transform.eulerAngles.y + 180) % 360;

            float faceToFactor = 0.0F;
            if (faceToAngle > 100 && faceToAngle < 260) faceToFactor = 0.25F;
            else if (faceToAngle < 80 || faceToAngle > 280) faceToFactor = 1.0F;
            else faceToFactor = 0.75F;

            detector.transform.LookAt(target.transform);
            float vAngle = Vector3.Angle(detector.transform.forward, focusCenter.transform.forward);
            float fofFactor = (float)Math.Pow(Mathf.Cos(Mathf.Clamp(Mathf.Min(fof, vAngle) * 90 / fof, 0, 89.99999F) * Mathf.PI / 180), 0.25F);
            float vDistance = Vector3.Distance(cameraBase.transform.position, target.transform.position);
            float thisRank = faceToFactor * fofFactor / vDistance;
            return thisRank;
        }
    }

    public class Anchors
    {
        public GameObject head;
        public GameObject chest;
        public GameObject hipF;
        public GameObject hipB;
        public FreeControllerV3 dancerHead;
        public FreeControllerV3 dancerChest;
        public FreeControllerV3 dancerHip;
        public FreeControllerV3 dancerRoot;
        public float rank = 1.0F;
        public float dist = 1.0F;
    }

    public class AutoFloat
    {
        public float timeRemain = 0.0F;
        public float targetVal = 0.0F;
        public float currentVal = 0.0F;
        public float deltaVal = 0.0F;
        public float baseValue = 0.0F;
        public bool isFiniched = true;
        public System.Random randomObj = new System.Random();
        public void SetNewAuto(float range, float timeMin, float timeMax)
        {
            targetVal = range * 2 * (float)randomObj.NextDouble() - range + baseValue;
            timeRemain = timeMin + (timeMax - timeMin) * (float)randomObj.NextDouble();
            deltaVal = (targetVal - currentVal) * Time.fixedDeltaTime / timeRemain;
            isFiniched = false;
        }
        public float GetNext()
        {
            if (isFiniched) return 0.0F;
            if (Mathf.Abs(targetVal - currentVal) <= Mathf.Abs(deltaVal) || timeRemain <= 0)
            {
                timeRemain = 0.0F;
                currentVal = targetVal;
                isFiniched = true;
            }
            else
            {
                timeRemain -= Time.fixedDeltaTime;
                currentVal += deltaVal;
            }
            return currentVal;
        }
    }
    public class V3Fellow
    {
        public Vector3 thisV3 = new Vector3(0.0F, 0.0F, 0.0F);
        public Vector3 targetV3 = new Vector3(0.0F, 0.0F, 0.0F);
        public Vector3 referenceV3 = new Vector3(0.0F, 0.0F, 0.0F);
        public float rateX = 1.0F;
        public float rateY = 1.0F;
        public float rateZ = 1.0F;
        public bool isAngle = false;
        public bool halfRound = false;
        public float factor = 0.0F;

        public Vector3 GetNewV3(Vector3 V3) { return GetNewVector3(NormalizeTarget(V3)); }

        public Vector3 NormalizeTarget(Vector3 V3)
        {
            Vector3 difV3 = V3 - referenceV3;
            if (isAngle)
            {
                difV3.x = (difV3.x + 180) % 360 - 180;
                difV3.y = (difV3.y + 180) % 360 - 180;
                difV3.z = (difV3.z + 180) % 360 - 180;
                if (halfRound)
                {
                    if (difV3.x > 90) difV3.x = 180 - difV3.x;
                    if (difV3.x < -90) difV3.x = -180 - difV3.x;
                    if (difV3.y > 90) difV3.y = 180 - difV3.y;
                    if (difV3.y < -90) difV3.y = -180 - difV3.y;
                    if (difV3.z > 90) difV3.z = 180 - difV3.z;
                    if (difV3.z < -90) difV3.z = -180 - difV3.z;
                }
            }

            difV3.x *= rateX;
            difV3.y *= rateY;
            difV3.z *= rateZ;
            targetV3 = difV3 + referenceV3;
            if (isAngle)
            {
                targetV3.x %= 360;
                targetV3.y %= 360;
                targetV3.z %= 360;
                if (targetV3.x > 90 && targetV3.x < 270)
                {
                    if (targetV3.x <= 180) targetV3.x = 90;
                    else targetV3.x = 270;
                }
            }
            return targetV3;
        }

        public Vector3 GetNewVector3(Vector3 V3)
        {
            Vector3 difV3 = V3 - thisV3;
            if (isAngle)
            {
                difV3.x = (difV3.x + 180) % 360 - 180;
                difV3.y = (difV3.y + 180) % 360 - 180;
                difV3.z = (difV3.z + 180) % 360 - 180;
            }
            thisV3 += difV3 * factor;

            if (isAngle)
            {
                thisV3.x %= 360;
                thisV3.y %= 360;
                thisV3.z %= 360;
                if (thisV3.x > 90 && thisV3.x < 270)
                {
                    if (thisV3.x <= 180) thisV3.x = 90;
                    else thisV3.x = 270;
                }
            }
            return thisV3;
        }

        public void SetThisV3(Vector3 V3) { thisV3 = V3; }

        public void SetRefV3(Vector3 V3) { referenceV3 = V3; }
    }
}